// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:matrix_sdk/src/event/event.dart';
import 'package:matrix_sdk/src/event/room/message_event.dart';
import 'package:matrix_sdk/src/event/room/state/state_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/store.dart';
import 'package:meta/meta.dart';
/// [Identifiable] objects that can possibly be stored.
abstract class Storables<I extends Id, T extends Identifiable<I>> {
  final Map<I, T> _memory = Map();

  final Store store;

  Storables({this.store});

  @protected
  Future<void> saveAll(Iterable<T> value);

  @protected
  Future<T> retrieve(I id);

  @protected
  Stream<T> retrieveAll();

  Future<T> operator [](I id) async {
    // If the room is in memory, return that one
    var fromMemory = _memory[id];
    if (fromMemory != null) {
      return fromMemory;
    }

    // Look in the store
    return await retrieve(id);
  }

  Future<void> addAll(Iterable<T> values) async {

    final nonNullValues = values.where((value) => value != null);
    // Add to memory
    for (final value in nonNullValues) {
      _memory[value.id] = value;
    }

    await saveAll(nonNullValues);
  }

  Stream<T> all() async* {
    for (var item in _memory.values) {
      yield item;
    }

    // Load all events to memory
    await for (var item in retrieveAll()) {
      bool shouldYield = false;

      // If the event was not already in memory, yield it
      _memory.putIfAbsent(item.id, () {
        shouldYield = true;
        return item;
      });

      if (shouldYield) {
        yield item;
      }
    }
  }
}
